/*
 *	\
 *
 *		1994.5.12	Copyright T.Kobayashi
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "strclass.h"
#include "parse.h"
#include "err.h"
#include "inlib.h"

int			ParseDebug ;
int			TopLevel;
int			MaxCodeSize = 64 * 1024 ;
int			MaxGlobals = 256 ;
int			MaxFunctions = 256 ;

static	void	ParseIf( CodeSentense* );
static	void	ParseWhile( CodeSentense* );
static	void	ParseDoWhile( CodeSentense* );
static	void	ParseFor( CodeSentense* );
static	void	ParseComplex( CodeSentense* );
static	void	ParseBreak( CodeSentense* );
static	void	ParseContinue( CodeSentense* );
static	void	ParseSentenseExp( void );

/*	̍\	*/
int		ParseSentense()
{
	int		ident, flag ;
	CodeSentense	*code ;

#if	PDEBUG
	if ( ParseDebug )
	{
		if ( Input.token == SYM_EOF )
			printf( "EOF\n" );
		else
			printf( "PARSE [%s:%4d][%d:%d] %s\n",
				InputFile, InputLine,
				Input.token >> 16, Input.token & 0xFFFF, Input.str );
	}
#endif

	/*	t@CI	*/
	if ( Input.token == SYM_EOF )
	{
/*
		FileClose();
		IdentPrivate( IdentGlobalVar );
		IdentPrivate( IdentFunction );
*/
		return FALSE ;
	}

	code = CodeSetSentense( InputFile, InputLine );

	switch( Input.token )
	{
		/*	錾	*/
		case SYM_VAR:
			ParseVar( code, TopLevel );
			break ;
		case SYM_FUNCTION:
			if ( TopLevel == FALSE )
				ParseError( "֐錾̈ʒusł" );
			TopLevel = FALSE ;
			ParseDefineFunction( code );
			TopLevel = TRUE ;
			break ;

		/*	s	*/
		case SYM_IF:
			ParseIf( code );
			break ;
		case SYM_WHILE:
			ParseWhile( code );
			break ;
		case SYM_DO:
			ParseDoWhile( code );
			break ;
		case SYM_FOR:
			ParseFor( code );
			break ;
		case SYM_BREAK:
			ParseBreak( code );
			break ;
		case SYM_CONTINUE:
			ParseContinue( code );
			break ;
		case SYM_RETURN:
			ParseReturn( code );
			break ;
		case SYM_OPEN2:		/*		*/
			ParseComplex( code );
			break ;
		case SYM_IDENT:		/*	܂͎	*/
			if ( ( ident = IdentSearch( IdentLocalVar, Input.str ) ) >= 0 )
			{
				ParseLocalVar( ident );
				flag = IdentFlag( IdentLocalVar, ident );
				if ( flag & IDENT_CONST )
					ParseError( "Ǐ萔 %s ɑ悤Ƃ܂B", IdentLocalVar->name[ident] );
				ParseAssign( code );
			}
			else if ( ( ident = IdentSearch( IdentGlobalVar, Input.str ) ) >= 0 )
			{
				ParseGlobalVar( ident );
				flag = IdentFlag( IdentGlobalVar, ident );
				if ( flag & IDENT_CONST )
					ParseError( "萔 %s ɑ悤Ƃ܂B", IdentGlobalVar->name[ident] );
				ParseAssign( code );
			}
			else
			{
				code->stype = SENT_EXPRESS ;
				ParseSentenseExp();
				break ;
			}
			if ( flag & IDENT_CONST )
				ParseError( "萔ɑ悤Ƃ܂B");
			break ;
		case SYM_SEMCLN:	/*		*/
			code->stype = SENT_NULL ;
			GetToken();
			break ;

		default:
			code->stype = SENT_EXPRESS ;
			ParseSentenseExp();
	}
	code->next = CodeNextPtr();
	return TRUE ;
}

/*
 *		if ̍\
 */
static	void	ParseIf( code )
CodeSentense	*code ;
{
	GetToken();
	if ( Input.token != SYM_OPEN1 )
		ParseError( " ( KvłB" );
	else
		GetToken();

	/*		*/
	ParseExpression();
	CodeSetOperator( OPE_EXP_END );
	if ( Input.token != SYM_CLOSE1 )
		ParseError( " ) KvłB" );
	else
		GetToken();

	/*		*/
	ParseSentense();

	/*	else	*/
	if ( Input.token == SYM_ELSE )
	{
		code->stype = SENT_IF_ELSE ;
		GetToken();
		ParseSentense();
	}
	else
		code->stype = SENT_IF ;
}

/*
 *		while ̍\
 */
static	void	ParseWhile( code )
CodeSentense	*code ;
{
	code->stype = SENT_WHILE ;

	GetToken();
	if ( Input.token != SYM_OPEN1 )
		ParseError( " ( KvłB" );
	GetToken();

	/*		*/
	ParseExpression();
	CodeSetOperator( OPE_EXP_END );
	if ( Input.token != SYM_CLOSE1 )
		ParseError( " ) KvłB" );
	GetToken();

	/*		*/
	ParseSentense();
}

/*
 *		do-while ̍\
 */
static	void	ParseDoWhile( code )
CodeSentense	*code ;
{
	CodeConst	*adr ;

	code->stype = SENT_DO ;
	GetToken();

	/*	IAhX	*/
	adr = CodeSetInt( 0 );

	/*		*/
	ParseSentense();

	if ( Input.token != SYM_WHILE )
		ParseError( "while KvłB" );
	GetToken();
	if ( Input.token != SYM_OPEN1 )
		ParseError( " ( KvłB" );
	GetToken();

	/*		*/
	ParseExpression();
	CodeSetOperator( OPE_EXP_END );
	if ( Input.token != SYM_CLOSE1 )
		ParseError( " ) KvłB" );
	GetToken();
	if ( Input.token != SYM_SEMCLN )
		ParseError( " ; KvłB" );
	GetToken();

	adr->data.id.i = (int)CodeNextPtr();
}

/*
 *		for ̍\
 */
static	void	ParseFor( code )
CodeSentense	*code ;
{
	code->stype = SENT_FOR ;

	GetToken();
	if ( Input.token != SYM_OPEN1 )
		ParseError( " ( KvłB" );
	else
		GetToken();

	/*		*/
	ParseSentense();

	/*		*/
	ParseExpression();
	CodeSetOperator( OPE_EXP_END );
	if ( Input.token != SYM_SEMCLN )
		ParseError( " ; KvłB" );
	else
		GetToken();

	/*		*/
	ParseSentense();

	/*	{	*/
	ParseSentense();
}

/*
 *		̍\
 */
static	void	ParseComplex( code )
CodeSentense	*code ;
{
	CodeConst	*adr ;

#if	PDEBUG
	if ( ParseDebug )
		printf( "   Complex\n" );
#endif

	code->stype = SENT_COMPLEX ;
	GetToken();

	/*	IAhX	*/
	adr = CodeSetInt( 0 );

	while( Input.token != SYM_CLOSE2 )
	{
		if ( ParseSentense() == FALSE )
			ParseFatal( "Ńt@CGhɂȂ܂" );
	}
	GetToken();

	adr->data.id.i = (int)CodeNextPtr();
}

/*
 *		break ̍\
 */
static	void	ParseBreak( code )
CodeSentense	*code ;
{
	code->stype = SENT_BREAK ;
	GetToken();
	if ( Input.token != SYM_SEMCLN )
		ParseError( " ; KvłB" );
	else
		GetToken();
}

/*
 *		continue ̍\
 */
static	void	ParseContinue( code )
CodeSentense	*code ;
{
	code->stype = SENT_CONTINUE ;
	GetToken();
	if ( Input.token != SYM_SEMCLN )
		ParseError( " ; KvłB" );
	else
		GetToken();
}

/*	𕶂ɓo^	*/
static	void	ParseSentenseExp()
{
	ParseExpression();
	CodeSetOperator( OPE_EXP_END );
	if ( Input.token != SYM_SEMCLN )
		ParseError( ";KvłB" );
	GetToken();
}

